import mimetypes
import os
from collections import defaultdict
from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union, Any

from docarray.math.ndarray import check_arraylike_equality

if TYPE_CHECKING:  # pragma: no cover
    from docarray.score import NamedScore
    from docarray import DocumentArray, Document
    from docarray.typing import ArrayType, StructValueType, DocumentContentType

default_values = dict(
    granularity=0,
    adjacency=0,
    parent_id='',
    blob=b'',
    text='',
    weight=0.0,
    uri='',
    mime_type='',
    tags=dict,
    _metadata=dict,
    offset=0.0,
    location=list,
    modality='',
    evaluations='Dict[str, NamedScore]',
    scores='Dict[str, NamedScore]',
    chunks='ChunkArray',
    matches='MatchArray',
    timestamps=dict,
)

_all_mime_types = set(mimetypes.types_map.values())


def _is_not_empty(attribute, value):
    if value is not None:
        if attribute not in default_values:
            return True
        else:
            dv = default_values[attribute]
            if dv in (
                'ChunkArray',
                'MatchArray',
                'DocumentArray',
                list,
                dict,
                'Dict[str, NamedScore]',
            ):
                if value:
                    return True
            elif value != dv:
                return True
    return False


@dataclass(unsafe_hash=True, eq=False)
class DocumentData:
    _reference_doc: 'Document' = field(hash=False, compare=False)
    id: str = field(default_factory=lambda: os.urandom(16).hex())
    parent_id: Optional[str] = None
    granularity: Optional[int] = None
    adjacency: Optional[int] = None
    blob: Optional[bytes] = None
    tensor: Optional['ArrayType'] = field(default=None, hash=False, compare=False)
    mime_type: Optional[str] = None  # must be put in front of `text` `content`
    text: Optional[str] = None
    content: Optional['DocumentContentType'] = None
    weight: Optional[float] = None
    uri: Optional[str] = None
    tags: Optional[Dict[str, Any]] = None
    _metadata: Optional[Dict[str, 'StructValueType']] = None
    offset: Optional[float] = None
    location: Optional[List[float]] = None
    embedding: Optional['ArrayType'] = field(default=None, hash=False, compare=False)
    modality: Optional[str] = None
    evaluations: Optional[Dict[str, Union['NamedScore', Dict]]] = None
    scores: Optional[Dict[str, Union['NamedScore', Dict]]] = None
    chunks: Optional['DocumentArray'] = None
    matches: Optional['DocumentArray'] = None

    @property
    def _non_empty_fields(self) -> Tuple[str]:
        r = []
        for f in fields(self):
            f_name = f.name
            if not f_name.startswith('_') or f_name == '_metadata':
                v = getattr(self, f_name)
                if _is_not_empty(f_name, v):
                    r.append(f_name)

        return tuple(r)

    def _set_default_value_if_none(self, key):
        if getattr(self, key) is None:
            v = default_values.get(key, None)
            if v is not None:
                if v == 'DocumentArray':
                    from docarray import DocumentArray

                    setattr(self, key, DocumentArray())
                elif v == 'ChunkArray':
                    from docarray.array.chunk import ChunkArray

                    setattr(
                        self, key, ChunkArray(None, reference_doc=self._reference_doc)
                    )
                elif v == 'MatchArray':
                    from docarray.array.match import MatchArray

                    setattr(
                        self, key, MatchArray(None, reference_doc=self._reference_doc)
                    )
                elif v == 'Dict[str, NamedScore]':
                    from docarray.score import NamedScore

                    setattr(self, key, defaultdict(NamedScore))
                else:
                    setattr(self, key, v() if callable(v) else v)

    @staticmethod
    def _embedding_eq(array1: 'ArrayType', array2: 'ArrayType'):

        if array1 is None and array2 is None:
            return True

        if type(array1) == type(array2):
            return check_arraylike_equality(array1, array2)
        else:
            return False

    @staticmethod
    def _tensor_eq(array1: 'ArrayType', array2: 'ArrayType'):
        DocumentData._embedding_eq(array1, array2)

    def __eq__(self, other):

        self_non_empty_fields = self._non_empty_fields
        other_non_empty_fields = other._non_empty_fields

        if other_non_empty_fields != self_non_empty_fields:
            return False

        for key in self_non_empty_fields:

            if hasattr(self, f'_{key}_eq'):

                if hasattr(DocumentData, f'_{key}_eq'):
                    are_equal = getattr(DocumentData, f'_{key}_eq')(
                        getattr(self, key), getattr(other, key)
                    )
                    print(
                        f'are_equal( {getattr(self, key)}, {getattr(other, key)}) ---> {are_equal}'
                    )
                    if are_equal == False:
                        return False
            else:
                if getattr(self, key) != getattr(other, key):
                    return False
        return True



import threading
import time
from typing import (
    Union,
    BinaryIO,
    TYPE_CHECKING,
    Generator,
    Type,
    Dict,
    Optional,
    Tuple,
)

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    from docarray.typing import T
    from docarray import Document


class VideoDataMixin:
    """Provide helper functions for :class:`Document` to support video data."""

    @classmethod
    def generator_from_webcam(
        cls: Type['T'],
        height_width: Optional[Tuple[int, int]] = None,
        show_window: bool = True,
        window_title: str = 'webcam',
        fps: int = 30,
        exit_key: int = 27,
        exit_event=None,
        tags: Optional[Dict] = None,
    ) -> Generator['T', None, None]:
        """
        Create a generator that yields a :class:`Document` object from the webcam.

        This feature requires the `opencv-python` package.

        :param height_width: the shape of the video frame, if not provided, the shape will be determined from the first frame.
            Note that this is restricted by the hardware of the camera.
        :param show_window: if to show preview window of the webcam video
        :param window_title: the window title of the preview window
        :param fps: expected frames per second, note that this is not guaranteed, as the actual fps depends on the hardware limit
        :param exit_key: the key to press to exit the preview window
        :param exit_event: the multiprocessing/threading/asyncio event that once set to exit the preview window
        :param tags: the tags to attach to the document
        :return: a generator that yields a :class:`Document` object from a webcam
        """
        import cv2

        if exit_event is None:
            exit_event = threading.Event()

        vc = cv2.VideoCapture(0)
        prev_frame_time = time.perf_counter()
        actual_fps = 0

        try:
            while not exit_event.is_set():
                rval, frame = vc.read()
                d = cls(tensor=frame, tags=tags)  # type: Document
                if height_width:
                    d.set_image_tensor_shape(height_width)

                yield d

                key = cv2.waitKey(1000 // (fps + fps - actual_fps))

                if show_window:
                    new_frame_time = time.perf_counter()

                    actual_fps = int(1 / (new_frame_time - prev_frame_time))
                    prev_frame_time = new_frame_time

                    # converting the fps into integer

                    # putting the FPS count on the frame
                    cv2.putText(
                        d.tensor,
                        f'FPS {actual_fps:0.0f}/{fps}',
                        (7, 70),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        3,
                        (255, 255, 255),
                        3,
                        cv2.LINE_AA,
                    )

                    # displaying the frame with fps
                    cv2.imshow(window_title, d.tensor)

                if key == exit_key or not rval:
                    break
        finally:
            vc.release()
            if show_window:
                cv2.destroyWindow(window_title)

    def load_uri_to_video_tensor(
        self: 'T', only_keyframes: bool = False, **kwargs
    ) -> 'T':
        """Convert a :attr:`.uri` to a video ndarray :attr:`.tensor`.

        :param only_keyframes: only keep the keyframes in the video
        :param kwargs: supports all keyword arguments that are being supported by av.open() as
            described in: https://pyav.org/docs/stable/api/_globals.html?highlight=open#av.open
        :return: Document itself after processed
        """
        import av

        with av.open(self.uri, **kwargs) as container:
            if only_keyframes:
                stream = container.streams.video[0]
                stream.codec_context.skip_frame = 'NONKEY'

            frames = []
            for frame in container.decode(video=0):
                img = frame.to_image()
                frames.append(np.asarray(img))

        self.tensor = np.moveaxis(np.stack(frames), 1, 2)
        return self

    def save_video_tensor_to_file(
        self: 'T', file: Union[str, BinaryIO], frame_rate: int = 30, codec: str = 'h264'
    ) -> 'T':
        """Save :attr:`.tensor` as a video mp4/h264 file.

        :param file: The file to open, which can be either a string or a file-like object.
        :param frame_rate: frames per second
        :param codec: the name of a decoder/encoder
        :return: itself after processed
        """
        if (
            self.tensor.ndim != 4
            or self.tensor.shape[-1] != 3
            or self.tensor.dtype != np.uint8
        ):
            raise ValueError(
                f'expects `.tensor` with dtype=uint8 and ndim=4 and the last dimension is 3, '
                f'but receiving {self.tensor.shape} in {self.tensor.dtype}'
            )

        video_tensor = np.moveaxis(np.clip(self.tensor, 0, 255), 1, 2)

        import av

        with av.open(file, mode='w') as container:
            stream = container.add_stream(codec, rate=frame_rate)
            stream.width = self.tensor.shape[1]
            stream.height = self.tensor.shape[2]
            stream.pix_fmt = 'yuv420p'

            for b in video_tensor:
                frame = av.VideoFrame.from_ndarray(b, format='rgb24')
                for packet in stream.encode(frame):
                    container.mux(packet)

            for packet in stream.encode():
                container.mux(packet)
        return self



from typing import TYPE_CHECKING, Union

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    from docarray.typing import T
    import trimesh


class Mesh:
    FILE_EXTENSIONS = [
        'glb',
        'obj',
        'ply',
    ]
    VERTICES = 'vertices'
    FACES = 'faces'


class MeshDataMixin:
    """Provide helper functions for :class:`Document` to support 3D mesh data and point cloud."""

    def _load_mesh(
        self, force: str = None
    ) -> Union['trimesh.Trimesh', 'trimesh.Scene']:
        """Load a trimesh.Mesh or trimesh.Scene object from :attr:`.uri`.

        :param force: str or None. For 'mesh' try to coerce scenes into a single mesh. For 'scene'
            try to coerce everything into a scene.
        :return: trimesh.Mesh or trimesh.Scene object
        """
        import urllib.parse
        import trimesh

        scheme = urllib.parse.urlparse(self.uri).scheme
        loader = trimesh.load_remote if scheme in ['http', 'https'] else trimesh.load

        mesh = loader(self.uri, force=force)

        return mesh

    def load_uri_to_point_cloud_tensor(
        self: 'T', samples: int, as_chunks: bool = False
    ) -> 'T':
        """Convert a 3d mesh-like :attr:`.uri` into :attr:`.tensor`

        :param samples: number of points to sample from the mesh
        :param as_chunks: when multiple geometry stored in one mesh file,
            then store each geometry into different :attr:`.chunks`

        :return: itself after processed
        """

        if as_chunks:
            import trimesh
            from docarray.document import Document

            # try to coerce everything into a scene
            scene = self._load_mesh(force='scene')
            for geo in scene.geometry.values():
                geo: trimesh.Trimesh
                self.chunks.append(Document(tensor=np.array(geo.sample(samples))))
        else:
            # combine a scene into a single mesh
            mesh = self._load_mesh(force='mesh')
            self.tensor = np.array(mesh.sample(samples))

        return self

    def load_uri_to_vertices_and_faces(self: 'T') -> 'T':
        """Convert a 3d mesh-like :attr:`.uri` into :attr:`.chunks` as vertices and faces

        :return: itself after processed
        """
        from docarray.document import Document

        mesh = self._load_mesh(force='mesh')

        vertices = mesh.vertices.view(np.ndarray)
        faces = mesh.faces.view(np.ndarray)

        self.chunks = [
            Document(name=Mesh.VERTICES, tensor=vertices),
            Document(name=Mesh.FACES, tensor=faces),
        ]

        return self

    def load_vertices_and_faces_to_point_cloud(self: 'T', samples: int) -> 'T':
        """Convert a 3d mesh of vertices and faces from :attr:`.chunks` into point cloud :attr:`.tensor`

        :param samples: number of points to sample from the mesh
        :return: itself after processed
        """
        vertices = None
        faces = None

        for chunk in self.chunks:
            if chunk.tags['name'] == Mesh.VERTICES:
                vertices = chunk.tensor
            if chunk.tags['name'] == Mesh.FACES:
                faces = chunk.tensor

        if vertices is not None and faces is not None:
            import trimesh

            mesh = trimesh.Trimesh(vertices=vertices, faces=faces)
            self.tensor = np.array(mesh.sample(samples))
        else:
            raise AttributeError(
                'Point cloud tensor can not be set, since vertices and faces chunk tensor have not been set.'
            )

        return self

    def load_uris_to_rgbd_tensor(self: 'T') -> 'T':
        """Load RGB image from :attr:`.uri` of :attr:`.chunks[0]` and depth image from :attr:`.uri` of :attr:`.chunks[1]` and merge them into :attr:`.tensor`.

        :return: itself after processed
        """
        from PIL import Image

        if len(self.chunks) != 2:
            raise ValueError(
                f'The provided Document does not have two chunks but instead {len(self.chunks)}. To load uris to RGBD tensor, the Document needs to have two chunks, with the first one providing the RGB image uri, and the second one providing the depth image uri.'
            )
        for chunk in self.chunks:
            if chunk.uri == '':
                raise ValueError(
                    'A chunk of the given Document does not provide a uri.'
                )

        rgb_img = np.array(Image.open(self.chunks[0].uri).convert('RGB'))
        depth_img = np.array(Image.open(self.chunks[1].uri))

        if rgb_img.shape[0:2] != depth_img.shape:
            raise ValueError(
                f'The provided RGB image and depth image are not of the same shapes: {rgb_img.shape[0:2]} != {depth_img.shape}'
            )

        self.tensor = np.concatenate(
            (rgb_img, np.expand_dims(depth_img, axis=2)), axis=-1
        )

        return self



import json
import os
import os.path
import warnings
from pathlib import Path
from typing import Dict, Type, TYPE_CHECKING, List, Optional, Any

import hubble
from hubble import Client as HubbleClient
from hubble.client.endpoints import EndpointsV2

from docarray.helper import (
    __cache_path__,
    _get_array_info,
    get_full_version,
)

if TYPE_CHECKING:  # pragma: no cover
    from docarray.typing import T


def _get_length_from_summary(summary: List[Dict]) -> Optional[int]:
    """Get the length from summary."""
    for item in summary:
        if 'Length' == item['name']:
            return item['value']


class PushPullMixin:
    """Transmitting :class:`DocumentArray` via Jina Cloud Service"""

    _max_bytes = 4 * 1024 * 1024 * 1024

    @staticmethod
    @hubble.login_required
    def cloud_list(show_table: bool = False) -> List[str]:
        """List all available arrays in the cloud.

        :param show_table: if true, show the table of the arrays.
        :returns: List of available DocumentArray's names.
        """
        from rich import print

        result = []
        from rich.table import Table
        from rich import box

        resp = HubbleClient(jsonify=True).list_artifacts(
            filter={'type': 'documentArray'}, sort={'createdAt': 1}
        )

        table = Table(
            title=f'You have {resp["meta"]["total"]} DocumentArray on the cloud',
            box=box.SIMPLE,
            highlight=True,
        )
        table.add_column('Name')
        table.add_column('Length')
        table.add_column('Access')
        table.add_column('Created at', justify='center')
        table.add_column('Updated at', justify='center')

        for da in resp['data']:
            result.append(da['name'])

            table.add_row(
                da['name'],
                str(_get_length_from_summary(da['metaData'].get('summary', []))),
                da['visibility'],
                da['createdAt'],
                da['updatedAt'],
            )

        if show_table:
            print(table)
        return result

    @staticmethod
    @hubble.login_required
    def cloud_delete(name: str) -> None:
        """
        Delete a DocumentArray from the cloud.
        :param name: the name of the DocumentArray to delete.
        """
        HubbleClient(jsonify=True).delete_artifact(name=name)

    def _get_raw_summary(self) -> List[Dict[str, Any]]:
        (
            is_homo,
            _nested_in,
            _nested_items,
            attr_counter,
            all_attrs_names,
        ) = _get_array_info(self)

        items = [
            dict(
                name='Type',
                value=self.__class__.__name__,
                description='The type of the DocumentArray',
            ),
            dict(
                name='Length',
                value=len(self),
                description='The length of the DocumentArray',
            ),
            dict(
                name='Homogenous Documents',
                value=is_homo,
                description='Whether all documents are of the same structure, attributes',
            ),
            dict(
                name='Common Attributes',
                value=list(attr_counter.items())[0][0] if attr_counter else None,
                description='The common attributes of all documents',
            ),
            dict(
                name='Has nested Documents in',
                value=tuple(_nested_in),
                description='The field that contains nested Documents',
            ),
            dict(
                name='Multimodal dataclass',
                value=all(d.is_multimodal for d in self),
                description='Whether all documents are multimodal',
            ),
            dict(
                name='Subindices', value=tuple(getattr(self, '_subindices', {}).keys())
            ),
        ]

        items.append(
            dict(
                name='Inspect attributes',
                value=_nested_items,
                description='Quick overview of attributes of all documents',
            )
        )

        storage_infos = self._get_storage_infos()
        _nested_items = []
        if storage_infos:
            for k, v in storage_infos.items():
                _nested_items.append(dict(name=k, value=v))
        items.append(
            dict(
                name='Storage backend',
                value=_nested_items,
                description='Quick overview of the Document Store',
            )
        )

        return items

    @hubble.login_required
    def push(
        self,
        name: str,
        show_progress: bool = False,
        public: bool = True,
        branding: Optional[Dict] = None,
    ) -> Dict:
        """Push this DocumentArray object to Jina Cloud which can be later retrieved via :meth:`.push`

        .. note::
            - Push with the same ``name`` will override the existing content.
            - Kinda like a public clipboard where everyone can override anyone's content.
              So to make your content survive longer, you may want to use longer & more complicated name.
            - The lifetime of the content is not promised atm, could be a day, could be a week. Do not use it for
              persistence. Only use this full temporary transmission/storage/clipboard.

        :param name: a name that later can be used for retrieve this :class:`DocumentArray`.
        :param show_progress: if to show a progress bar on pulling
        :param public: by default anyone can pull a DocumentArray if they know its name.
            Setting this to False will allow only the creator to pull it. This feature of course you to login first.
        :param branding: a dict of branding information to be sent to Jina Cloud. {"icon": "emoji", "background": "#fff"}
        """
        import requests

        delimiter = os.urandom(32)

        (data, ctype) = requests.packages.urllib3.filepost.encode_multipart_formdata(
            {
                'file': (
                    'DocumentArray',
                    delimiter,
                ),
                'name': name,
                'type': 'documentArray',
                'public': public,
                'metaData': json.dumps(
                    {
                        'summary': self._get_raw_summary(),
                        'branding': branding,
                        'version': get_full_version(),
                    },
                    sort_keys=True,
                ),
            }
        )

        headers = {
            'Content-Type': ctype,
        }

        auth_token = hubble.get_token()
        if auth_token:
            headers['Authorization'] = f'token {auth_token}'

        _head, _tail = data.split(delimiter)
        _head += self._stream_header
        from rich import filesize
        from docarray.array.mixins.io.pbar import get_progressbar

        pbar, t = get_progressbar('Pushing', disable=not show_progress, total=len(self))

        def gen():
            total_size = 0

            pbar.start_task(t)

            yield _head

            def _get_chunk(_batch):
                return b''.join(
                    d._to_stream_bytes(protocol='protobuf', compress='gzip')
                    for d in _batch
                ), len(_batch)

            for chunk, num_doc_in_chunk in self.map_batch(_get_chunk, batch_size=32):
                total_size += len(chunk)
                if total_size > self._max_bytes:
                    warnings.warn(
                        f'DocumentArray is too big. The pushed DocumentArray might be chopped off.'
                    )
                    break
                yield chunk
                pbar.update(
                    t,
                    advance=num_doc_in_chunk,
                    total_size=str(filesize.decimal(total_size)),
                )
            yield _tail

        with pbar:

            response = requests.post(
                HubbleClient()._base_url + EndpointsV2.upload_artifact,
                data=gen(),
                headers=headers,
            )

        if response.ok:
            return response.json()['data']
        else:
            if response.status_code >= 400 and 'readableMessage' in response.json():
                response.reason = response.json()['readableMessage']
            response.raise_for_status()

    @classmethod
    @hubble.login_required
    def pull(
        cls: Type['T'],
        name: str,
        show_progress: bool = False,
        local_cache: bool = True,
        *args,
        **kwargs,
    ) -> 'T':
        """Pulling a :class:`DocumentArray` from Jina Cloud Service to local.

        :param name: the upload name set during :meth:`.push`
        :param show_progress: if to show a progress bar on pulling
        :param local_cache: store the downloaded DocumentArray to local folder
        :return: a :class:`DocumentArray` object
        """

        import requests

        headers = {}

        auth_token = hubble.get_token()

        if auth_token:
            headers['Authorization'] = f'token {auth_token}'

        url = HubbleClient()._base_url + EndpointsV2.download_artifact + f'?name={name}'
        response = requests.get(url, headers=headers)

        if response.ok:
            url = response.json()['data']['download']
        else:
            response.raise_for_status()

        with requests.get(
            url,
            stream=True,
        ) as r:
            r.raise_for_status()

            _da_len = int(r.headers['Content-length'])

            from docarray.array.mixins.io.binary import LazyRequestReader

            _source = LazyRequestReader(r)
            if local_cache and os.path.exists(f'{__cache_path__}/{name}'):
                _cache_len = os.path.getsize(f'{__cache_path__}/{name}')
                if _cache_len == _da_len:
                    _source = f'{__cache_path__}/{name}'

            r = cls.load_binary(
                _source,
                protocol='protobuf',
                compress='gzip',
                _show_progress=show_progress,
                *args,
                **kwargs,
            )

            if isinstance(_source, LazyRequestReader) and local_cache:
                Path(__cache_path__).mkdir(parents=True, exist_ok=True)
                with open(f'{__cache_path__}/{name}', 'wb') as fp:
                    fp.write(_source.content)

            return r

    cloud_push = push
    cloud_pull = pull



import warnings
from typing import Optional, Union, TYPE_CHECKING, Callable, List, Dict, Tuple

from functools import wraps

import numpy as np
from collections import defaultdict

from docarray.score import NamedScore

if TYPE_CHECKING:  # pragma: no cover
    from docarray import Document, DocumentArray
    from docarray.array.mixins.embed import CollateFnType
    from docarray.typing import ArrayType, AnyDNN


def _evaluate_deprecation(f):
    """Raises a deprecation warning if the user executes the evaluate function with
    the old interface and adjust the input to fit the new interface."""

    @wraps(f)
    def func(*args, **kwargs):
        if len(args) > 1:
            if not (
                isinstance(args[1], Callable)
                or isinstance(args[1], str)
                or isinstance(args[1], list)
            ):
                kwargs['ground_truth'] = args[1]
                args = [args[0]] + list(args[2:])
                warnings.warn(
                    'The `other` attribute in `evaluate()` is transfered from a '
                    'positional attribute into the keyword attribute `ground_truth`.'
                    'Using it as a positional attribute is deprecated and will be removed '
                    'soon.',
                    DeprecationWarning,
                )
        for old_key, new_key in zip(
            ['other', 'metric', 'metric_name'],
            ['ground_truth', 'metrics', 'metric_names'],
        ):
            if old_key in kwargs:
                kwargs[new_key] = kwargs[old_key]
                warnings.warn(
                    f'`{old_key}` is renamed to `{new_key}` in `evaluate()`, the '
                    f'usage of `{old_key}` is deprecated and will be removed soon.',
                    DeprecationWarning,
                )

        # transfer metrics and metric_names into lists
        list_warning_msg = (
            'The attribute `%s` now accepts a list instead of a '
            'single element. Passing a single element is deprecated and will soon not '
            'be supported anymore.'
        )
        if len(args) > 1:
            if type(args[1]) is str:
                args = list(args)
                args[1] = [args[1]]
                warnings.warn(list_warning_msg % 'metrics', DeprecationWarning)
        for key in ['metrics', 'metric_names']:
            if key in kwargs and type(kwargs[key]) is str:
                kwargs[key] = [kwargs[key]]
                warnings.warn(list_warning_msg % key, DeprecationWarning)
        return f(*args, **kwargs)

    return func


class EvaluationMixin:
    """A mixin that provides ranking evaluation functionality to DocumentArrayLike objects"""

    @_evaluate_deprecation
    def evaluate(
        self,
        metrics: List[Union[str, Callable[..., float]]],
        ground_truth: Optional['DocumentArray'] = None,
        hash_fn: Optional[Callable[['Document'], str]] = None,
        metric_names: Optional[List[str]] = None,
        strict: bool = True,
        label_tag: str = 'label',
        **kwargs,
    ) -> Dict[str, float]:
        """
        Compute ranking evaluation metrics for a given `DocumentArray` when compared
        with a ground truth.

        If one provides a `ground_truth` DocumentArray that is structurally identical
        to `self`, this function compares the `matches` of `documents` inside the
        `DocumentArray` to this `ground_truth`.
        Alternatively, one can directly annotate the documents by adding labels in the
        form of tags with the key specified in the `label_tag` attribute.
        Those tags need to be added to `self` as well as to the documents in the
        matches properties.

        This method will fill the `evaluations` field of Documents inside this
        `DocumentArray` and will return the average of the computations

        :param metrics: List of metric names or metric functions to be computed
        :param ground_truth: The ground_truth `DocumentArray` that the `DocumentArray`
            compares to.
        :param hash_fn: For the evaluation against a `ground_truth` DocumentArray,
            this function is used for generating hashes which are used to compare the
            documents. If not given, ``Document.id`` is used.
        :param metric_names: If provided, the results of the metrics computation will be
            stored in the `evaluations` field of each Document with this names. If not
            provided, the names will be derived from the metric function names.
        :param strict: If set, then left and right sides are required to be fully
            aligned: on the length, and on the semantic of length. These are preventing
            you to evaluate on irrelevant matches accidentally.
        :param label_tag: Specifies the tag which contains the labels.
        :param kwargs: Additional keyword arguments to be passed to the metric
            functions.
        :return: A dictionary which stores for each metric name the average evaluation
            score.
        """
        if len(self) == 0:
            raise ValueError('It is not possible to evaluate an empty DocumentArray')
        if ground_truth and len(ground_truth) > 0 and ground_truth[0].matches:
            ground_truth_type = 'matches'
        elif label_tag in self[0].tags:
            if ground_truth:
                warnings.warn(
                    'A ground_truth attribute is provided but does not '
                    'contain matches. The labels are used instead and '
                    'ground_truth is ignored.'
                )
            ground_truth = self
            ground_truth_type = 'labels'

        else:
            raise RuntimeError(
                'Could not find proper ground truth data. Either labels or the '
                'ground_truth attribute with matches is required'
            )

        if strict:
            self._check_length(len(ground_truth))

        if hash_fn is None:
            hash_fn = lambda d: d.id

        metric_fns = []
        for metric in metrics:
            if callable(metric):
                metric_fns.append(metric)
            elif isinstance(metric, str):
                from docarray.math import evaluation

                metric_fns.append(getattr(evaluation, metric))

        if not metric_names:
            metric_names = [metric_fn.__name__ for metric_fn in metric_fns]

        if len(metric_names) != len(metrics):
            raise ValueError(
                'Could not match metric names to the metrics since the number of '
                'metric names does not match the number of metrics'
            )

        results = defaultdict(list)
        caller_max_rel = kwargs.pop('max_rel', None)
        for d, gd in zip(self, ground_truth):
            max_rel = caller_max_rel or len(gd.matches)
            if strict and hash_fn(d) != hash_fn(gd):
                raise ValueError(
                    f'Document {d} from the left-hand side and '
                    f'{gd} from the right-hand are not hashed to the same value. '
                    f'This means your left and right DocumentArray may not be aligned; or it means your '
                    f'`hash_fn` is badly designed.'
                )
            if not d.matches or not gd.matches:
                raise ValueError(
                    f'Document {d!r} or {gd!r} has no matches, please check your Document'
                )

            targets = gd.matches[:max_rel]

            if ground_truth_type == 'matches':
                desired = {hash_fn(m) for m in targets}
                if len(desired) != len(targets):
                    warnings.warn(
                        f'{hash_fn!r} may not be valid, as it maps multiple Documents into the same hash. '
                        f'Evaluation results may be affected'
                    )
                binary_relevance = [
                    1 if hash_fn(m) in desired else 0 for m in d.matches
                ]
            elif ground_truth_type == 'labels':
                binary_relevance = [
                    1 if m.tags[label_tag] == d.tags[label_tag] else 0 for m in targets
                ]
            else:
                raise RuntimeError(
                    'Could not identify which kind of ground truth'
                    'information is provided to evaluate the matches.'
                )
            for metric_name, metric_fn in zip(metric_names, metric_fns):
                if 'max_rel' in metric_fn.__code__.co_varnames:
                    kwargs['max_rel'] = max_rel
                r = metric_fn(binary_relevance, **kwargs)
                d.evaluations[metric_name] = NamedScore(
                    value=r, op_name=str(metric_fn), ref_id=d.id
                )
                results[metric_name].append(r)
        return {
            metric_name: float(np.mean(values))
            for metric_name, values in results.items()
        }

    def embed_and_evaluate(
        self,
        metrics: List[Union[str, Callable[..., float]]],
        index_data: Optional['DocumentArray'] = None,
        ground_truth: Optional['DocumentArray'] = None,
        metric_names: Optional[str] = None,
        strict: bool = True,
        label_tag: str = 'label',
        embed_models: Optional[Union['AnyDNN', Tuple['AnyDNN', 'AnyDNN']]] = None,
        embed_funcs: Optional[Union[Callable, Tuple[Callable, Callable]]] = None,
        device: str = 'cpu',
        batch_size: Union[int, Tuple[int, int]] = 256,
        collate_fns: Union[
            Optional['CollateFnType'],
            Tuple[Optional['CollateFnType'], Optional['CollateFnType']],
        ] = None,
        distance: Union[
            str, Callable[['ArrayType', 'ArrayType'], 'np.ndarray']
        ] = 'cosine',
        limit: Optional[Union[int, float]] = 20,
        normalization: Optional[Tuple[float, float]] = None,
        exclude_self: bool = False,
        use_scipy: bool = False,
        num_worker: int = 1,
        match_batch_size: int = 100_000,
        query_sample_size: int = 1_000,
        **kwargs,
    ) -> Optional[Union[float, List[float]]]:  # average for each metric
        """
        Computes ranking evaluation metrics for a given `DocumentArray`. This
        function does embedding and matching in the same turn. Thus, you don't need to
        call ``embed`` and ``match`` before it. Instead, it embeds the documents in
        `self` (and `index_data` when provided`) and compute the nearest neighbour
        itself. This might be done in batches for the `index_data` object to reduce
        the memory consumption of the evlauation process. The evaluation itself can be
        done against a `ground_truth` DocumentArray or on the basis of labels like it
        is possible with the :func:``evaluate`` function.

        :param metrics: List of metric names or metric functions to be computed
        :param index_data: The other DocumentArray  to match against, if not given,
            `self` will be matched against itself. This means that every document in
            will be compared to all other documents in `self` to determine the nearest
            neighbors.
        :param ground_truth: The ground_truth `DocumentArray` that the `DocumentArray`
            compares to.
        :param metric_names: If provided, the results of the metrics computation will be
            stored in the `evaluations` field of each Document with this names. If not
            provided, the names will be derived from the metric function names.
        :param strict: If set, then left and right sides are required to be fully
            aligned: on the length, and on the semantic of length. These are preventing
            you to evaluate on irrelevant matches accidentally.
        :param label_tag: Specifies the tag which contains the labels.
        :param embed_models: One or two embedding model written in Keras / Pytorch /
            Paddle for embedding `self` and `index_data`.
        :param embed_funcs: As an alternative to embedding models, custom embedding
            functions can be provided.
        :param device: the computational device for `embed_models`, and the matching
            can be either `cpu` or `cuda`.
        :param batch_size: Number of documents in a batch for embedding.
        :param collate_fns: For each embedding function the respective collate
            function creates a mini-batch of input(s) from the given `DocumentArray`.
            If not provided a default built-in collate_fn uses the `tensors` of the
            documents to create input batches.
        :param distance: The distance metric.
        :param limit: The maximum number of matches, when not given defaults to 20.
        :param normalization: A tuple [a, b] to be used with min-max normalization,
            the min distance will be rescaled to `a`, the max distance will be
            rescaled to `b` all values will be rescaled into range `[a, b]`.
        :param exclude_self: If set, Documents in ``index_data`` with same ``id``
            as the left-hand values will not be considered as matches.
        :param use_scipy: if set, use ``scipy`` as the computation backend. Note,
            ``scipy`` does not support distance on sparse matrix.
        :param num_worker: Specifies the number of workers for the execution of the
            match function.
        :parma match_batch_size: The number of documents which are embedded and
            matched at once. Set this value to a lower value, if you experience high
            memory consumption.
        :param kwargs: Additional keyword arguments to be passed to the metric
            functions.
        :param query_sample_size: For a large number of documents in `self` the
            evaluation becomes infeasible, especially, if `index_data` is large.
            Therefore, queries are sampled if the number of documents in `self` exceeds
            `query_sample_size`. Usually, this has only small impact on the mean metric
            values returned by this function. To prevent sampling, you can set
            `query_sample_size` to None.
        :return: A dictionary which stores for each metric name the average evaluation
            score.
        """

        from docarray import Document, DocumentArray

        if not query_sample_size:
            query_sample_size = len(self)

        query_data = self
        only_one_dataset = not index_data
        apply_sampling = len(self) > query_sample_size

        if only_one_dataset:
            # if the user does not provide a separate set of documents for indexing,
            # the matching is done on the documents itself
            copy_flag = (
                apply_sampling
                or (type(embed_funcs) is tuple)
                or ((embed_funcs is None) and (type(embed_models) is tuple))
            )
            index_data = DocumentArray(self, copy=True) if copy_flag else self

        if apply_sampling:
            rng = np.random.default_rng()
            query_data = DocumentArray(
                rng.choice(self, size=query_sample_size, replace=False)
            )

        if ground_truth and apply_sampling:
            ground_truth = DocumentArray(
                [ground_truth[d.id] for d in query_data if d.id in ground_truth]
            )
            if len(ground_truth) != len(query_data):
                raise ValueError(
                    'The DocumentArray provided in the ground_truth attribute does '
                    'not contain all the documents in self.'
                )

        index_data_labels = None
        if not ground_truth:
            if not label_tag in query_data[0].tags:
                raise ValueError(
                    'Either a ground_truth `DocumentArray` or labels are '
                    'required for the evaluation.'
                )
            if not label_tag in index_data[0].tags:
                raise ValueError(
                    'The `DocumentArray` provided in `index_data` misses ' 'labels.'
                )
            index_data_labels = dict()
            for id_value, tags in zip(index_data[:, 'id'], index_data[:, 'tags']):
                index_data_labels[id_value] = tags[label_tag]

        if embed_funcs is None:
            # derive embed function from embed model
            if embed_models is None:
                raise RuntimeError(
                    'For embedding the documents you need to provide either embedding '
                    'model(s) or embedding function(s)'
                )
            else:
                if type(embed_models) is not tuple:
                    embed_models = (embed_models, embed_models)
                embed_args = [
                    {
                        'embed_model': model,
                        'device': device,
                        'batch_size': batch_size,
                        'collate_fn': collate_fns[i]
                        if type(collate_fns) is tuple
                        else collate_fns,
                    }
                    for i, (model, docs) in enumerate(
                        zip(embed_models, (query_data, index_data))
                    )
                ]
        else:
            if type(embed_funcs) is not tuple:
                embed_funcs = (
                    embed_funcs,
                    embed_funcs,
                )  # use the same embedding function for queries and index

        # embed queries:
        if embed_funcs:
            embed_funcs[0](query_data)
        else:
            query_data.embed(**embed_args[0])

        for doc in query_data:
            doc.matches.clear()

        local_queries = DocumentArray(
            [Document(id=doc.id, embedding=doc.embedding) for doc in query_data]
        )

        def fuse_matches(global_matches: DocumentArray, local_matches: DocumentArray):
            global_matches.extend(local_matches)
            global_matches = sorted(
                global_matches,
                key=lambda x: x.scores[distance].value,
            )[:limit]
            return DocumentArray(global_matches)

        for batch in index_data.batch(match_batch_size):
            if (
                apply_sampling
                or (batch.embeddings is None)
                or (batch[0].embedding[0] == 0)
            ):
                if embed_funcs:
                    embed_funcs[1](batch)
                else:
                    batch.embed(**embed_args[1])

            local_queries.match(
                batch,
                limit=limit,
                metric=distance,
                normalization=normalization,
                exclude_self=exclude_self,
                use_scipy=use_scipy,
                num_worker=num_worker,
                device=device,
                batch_size=int(len(batch) / num_worker) if num_worker > 1 else None,
                only_id=True,
            )

            for doc in local_queries:
                query_data[doc.id, 'matches'] = fuse_matches(
                    query_data[doc.id].matches,
                    doc.matches,
                )

            batch.embeddings = None
        # set labels if necessary
        if not ground_truth:
            for i, doc in enumerate(query_data):
                new_matches = DocumentArray()
                for m in doc.matches:
                    m.tags = {label_tag: index_data_labels[m.id]}
                    new_matches.append(m)
                query_data[doc.id, 'matches'] = new_matches

        metrics_resp = query_data.evaluate(
            ground_truth=ground_truth,
            metrics=metrics,
            metric_names=metric_names,
            strict=strict,
            label_tag=label_tag,
            **kwargs,
        )

        return metrics_resp



import warnings
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, TypeVar, Union

import numpy as np
from docarray import Document, DocumentArray
from docarray.array.mixins.find import FindMixin as BaseFindMixin
from docarray.math import ndarray
from docarray.math.ndarray import to_numpy_array
from docarray.score import NamedScore

from redis.commands.search.query import Query
from redis.commands.search.querystring import (
    DistjunctUnion,
    IntersectNode,
    equal,
    ge,
    gt,
    intersect,
    le,
    lt,
    union,
)

if TYPE_CHECKING:  # pragma: no cover
    import tensorflow
    import torch

    RedisArrayType = TypeVar(
        'RedisArrayType',
        np.ndarray,
        tensorflow.Tensor,
        torch.Tensor,
        Sequence[float],
        Dict,
    )


class FindMixin(BaseFindMixin):
    def _find_similar_vectors(
        self,
        query: 'RedisArrayType',
        filter: Optional[Union[str, Dict]] = None,
        limit: Union[int, float] = 20,
        **kwargs,
    ):

        if filter:
            query_str = _get_redis_filter_query(filter)
        else:
            query_str = '*'

        q = (
            Query(f'({query_str})=>[KNN {limit} @embedding $vec AS vector_score]')
            .sort_by('vector_score')
            .paging(0, limit)
            .dialect(2)
        )

        query_params = {'vec': to_numpy_array(query).astype(np.float32).tobytes()}
        results = (
            self._client.ft(index_name=self._config.index_name)
            .search(q, query_params)
            .docs
        )

        da = DocumentArray()
        for res in results:
            doc = Document.from_base64(res.blob.encode())
            doc.scores['score'] = NamedScore(value=float(res.vector_score))
            da.append(doc)
        return da

    def _find(
        self,
        query: 'RedisArrayType',
        limit: Union[int, float] = 20,
        filter: Optional[Union[str, Dict]] = None,
        **kwargs,
    ) -> List['DocumentArray']:

        query = np.array(query)
        num_rows, n_dim = ndarray.get_array_rows(query)
        if n_dim != 2:
            query = query.reshape((num_rows, -1))

        return [
            self._find_similar_vectors(q, filter=filter, limit=limit, **kwargs)
            for q in query
        ]

    def _find_with_filter(
        self,
        filter: Union[str, Dict],
        limit: Union[int, float] = 20,
    ):
        query_str = _get_redis_filter_query(filter)
        q = Query(query_str)
        q.paging(0, limit)

        results = self._client.ft(index_name=self._config.index_name).search(q).docs

        da = DocumentArray()
        for res in results:
            doc = Document.from_base64(res.blob.encode())
            da.append(doc)
        return da

    def _filter(
        self,
        filter: Union[str, Dict],
        limit: Union[int, float] = 20,
    ) -> 'DocumentArray':

        return self._find_with_filter(filter, limit=limit)

    def _find_by_text(
        self,
        query: Union[str, List[str]],
        index: str = 'text',
        filter: Optional[Union[str, Dict]] = None,
        limit: Union[int, float] = 20,
        **kwargs,
    ):
        if isinstance(query, str):
            query = [query]

        return [
            self._find_similar_documents_from_text(
                q,
                index=index,
                filter=filter,
                limit=limit,
                **kwargs,
            )
            for q in query
        ]

    def _find_similar_documents_from_text(
        self,
        query: str,
        index: str = 'text',
        filter: Optional[Union[str, Dict]] = None,
        limit: Union[int, float] = 20,
        **kwargs,
    ):
        query_str = _build_query_str(query)

        if filter:
            filter_str = _get_redis_filter_query(filter)
        else:
            filter_str = ''

        scorer = kwargs.get('scorer', 'BM25')
        if scorer not in [
            'BM25',
            'TFIDF',
            'TFIDF.DOCNORM',
            'DISMAX',
            'DOCSCORE',
            'HAMMING',
        ]:
            raise ValueError(
                f'Expecting a valid text similarity ranking algorithm, got {scorer} instead'
            )

        q = Query(f'@{index}:{query_str} {filter_str}').scorer(scorer).paging(0, limit)

        results = self._client.ft(index_name=self._config.index_name).search(q).docs

        da = DocumentArray()
        for res in results:
            doc = Document.from_base64(res.blob.encode())
            da.append(doc)
        return da


def _build_query_node(key, condition):
    operator = list(condition.keys())[0]
    value = condition[operator]

    query_dict = {}

    if operator in ['$ne', '$eq']:
        if isinstance(value, bool):
            query_dict[key] = equal(int(value))
        elif isinstance(value, (int, float)):
            query_dict[key] = equal(value)
        else:
            query_dict[key] = value
    elif operator == '$gt':
        query_dict[key] = gt(value)
    elif operator == '$gte':
        query_dict[key] = ge(value)
    elif operator == '$lt':
        query_dict[key] = lt(value)
    elif operator == '$lte':
        query_dict[key] = le(value)
    else:
        raise ValueError(
            f'Expecting filter operator one of $gt, $gte, $lt, $lte, $eq, $ne, $and OR $or, got {operator} instead'
        )

    if operator == '$ne':
        return DistjunctUnion(**query_dict)
    return IntersectNode(**query_dict)


def _build_query_nodes(filter):
    nodes = []
    for k, v in filter.items():
        if k == '$and':
            children = _build_query_nodes(v)
            node = intersect(*children)
            nodes.append(node)
        elif k == '$or':
            children = _build_query_nodes(v)
            node = union(*children)
            nodes.append(node)
        else:
            child = _build_query_node(k, v)
            nodes.append(child)

    return nodes


def _build_query_str(query):
    query_str = '|'.join(query.split(' '))
    return query_str


def _get_redis_filter_query(filter: Union[str, Dict]):
    if isinstance(filter, dict):
        warnings.warn(
            "Dict syntax for redis filter will be deprecated, use string literals instead",
            DeprecationWarning,
        )
        nodes = _build_query_nodes(filter)
        query_str = intersect(*nodes).to_string()
    elif isinstance(filter, str):
        query_str = filter
    else:
        raise ValueError(f'Unexpected type of filter: {type(filter)}, expected str')

    return query_str



from typing import TYPE_CHECKING
from docarray.dataclasses.enums import DocumentMetadata, ImageType

if TYPE_CHECKING:  # pragma: no cover
    from docarray import Document


def image_getter(doc: 'Document'):
    if doc._metadata[DocumentMetadata.IMAGE_TYPE] == ImageType.URI:
        return doc.uri
    elif doc._metadata[DocumentMetadata.IMAGE_TYPE] == ImageType.PIL:
        from PIL import Image

        return Image.fromarray(doc.tensor)
    elif doc._metadata[DocumentMetadata.IMAGE_TYPE] == ImageType.NDARRAY:
        return doc.tensor


def text_getter(doc: 'Document'):
    return doc.text


def uri_getter(doc: 'Document'):
    return doc.uri


def audio_getter(doc: 'Document'):
    return doc.uri or doc.tensor


def video_getter(doc: 'Document'):
    return doc.uri or doc.tensor


def mesh_getter(doc: 'Document'):
    return doc.uri or doc.tensor


def tabular_getter(doc: 'Document'):
    return doc.uri


def blob_getter(doc: 'Document'):
    return doc.uri or doc.blob


def json_getter(doc: 'Document'):
    return doc.tags



from typing import Iterable, Dict

from docarray.array.storage.base.getsetdel import BaseGetSetDelMixin
from docarray.array.storage.base.helper import Offset2ID
from docarray import Document


class GetSetDelMixin(BaseGetSetDelMixin):
    """Provide concrete implementation for ``__getitem__``, ``__setitem__``,
    and ``__delitem__`` for ``DocumentArrayWeaviate``"""

    def _getitem(self, wid: str) -> 'Document':
        """Helper method for getting item with weaviate as storage

        :param wid: weaviate id
        :raises KeyError: raise error when weaviate id does not exist in storage
        :return: Document
        """
        try:
            resp = self._client.data_object.get_by_id(
                wid, with_vector=True, class_name=self._class_name
            )
            return Document.from_base64(
                resp['properties']['_serialized'], **self._serialize_config
            )
        except Exception as ex:
            raise KeyError(wid) from ex

    def _get_doc_by_id(self, _id: str) -> 'Document':
        """Concrete implementation of base class' ``_get_doc_by_id``

        :param _id: the id of the document
        :return: the retrieved document from weaviate
        """
        return self._getitem(self._map_id(_id))

    def _set_doc_by_id(self, _id: str, value: 'Document', flush: bool = True):
        """Concrete implementation of base class' ``_set_doc_by_id``

        :param _id: the id of doc to update
        :param value: the document to update to
        """
        if _id != value.id:
            self._del_doc_by_id(_id)

        payload = self._doc2weaviate_create_payload(value)
        self._client.batch.add_data_object(**payload)
        if flush:
            self._client.batch.flush()

    def _set_docs_by_ids(self, ids, docs: Iterable['Document'], mismatch_ids: Dict):
        """Overridden implementation of _set_docs_by_ids in order to add docs in batches and flush at the end

        :param ids: the ids used for indexing
        """
        for _id, doc in zip(ids, docs):
            self._set_doc_by_id(_id, doc, flush=False)
        self._client.batch.flush()

    def _del_doc_by_id(self, _id: str):
        """Concrete implementation of base class' ``_del_doc_by_id``

        :param _id: the id of the document to delete
        """
        if self._client.data_object.exists(
            self._map_id(_id), class_name=self._class_name
        ):
            self._client.data_object.delete(
                self._map_id(_id), class_name=self._class_name
            )

    def _clear_storage(self):
        """Concrete implementation of base class' ``_clear_storage``"""
        if self._class_name:
            self._client.schema.delete_class(self._class_name)
            self._client.schema.delete_class(self._meta_name)
            self._load_or_create_weaviate_schema()

    def _load_offset2ids(self):
        if self._list_like:
            ids, self._offset2ids_wid = self._get_offset2ids_meta()
            self._offset2ids = Offset2ID(ids, list_like=self._list_like)
        else:
            self._offset2ids = Offset2ID([], list_like=self._list_like)

    def _save_offset2ids(self):
        if self._list_like:
            self._update_offset2ids_meta()



from typing import Iterable, Union

from docarray import Document, DocumentArray
from docarray.array.storage.base.seqlike import BaseSequenceLikeMixin


class SequenceLikeMixin(BaseSequenceLikeMixin):
    """Implement sequence-like methods for DocumentArray with Redis as storage"""

    def __eq__(self, other):
        """Compare this object to the other, returns True if and only if other
        as the same type as self and other has the same meta information

        :param other: the other object to check for equality
        :return: ``True`` if other is equal to self
        """
        # two DA are considered as the same if they have the same client meta data
        return (
            type(self) is type(other)
            and self._client.client_info() == other._client.client_info()
            and self._config == other._config
        )

    def __len__(self):
        """Return the length of :class:`DocumentArray` that uses Redis as storage

        :return: the length of this :class:`DocumentArrayRedis` object
        """
        try:
            return len(self._offset2ids)
        except:
            return 0

    def __contains__(self, x: Union[str, 'Document']):
        """Check if ``x`` is contained in this :class:`DocumentArray` with Redis storage

        :param x: the id of the document to check or the document object itself
        :return: True if ``x`` is contained in self
        """
        if isinstance(x, str):
            return self._doc_id_exists(x)
        elif isinstance(x, Document):
            return self._doc_id_exists(x.id)
        else:
            return False

    def __repr__(self):
        """Return the string representation of :class:`DocumentArrayRedis` object
        :return: string representation of this object
        """
        return f'<DocumentArray[Redis] (length={len(self)}) at {id(self)}>'

    def _upload_batch(self, batch_of_docs: Iterable['Document']):
        pipe = self._client.pipeline()
        for doc in batch_of_docs:
            payload = self._document_to_redis(doc)
            pipe.hset(self._doc_prefix + doc.id, mapping=payload)
        pipe.execute()

    def _extend(self, docs: Iterable['Document']):
        da = DocumentArray(docs)
        for batch_of_docs in da.batch(self._config.batch_size):
            self._upload_batch(batch_of_docs)
            if self._list_like:
                self._offset2ids.extend(batch_of_docs[:, 'id'])



from abc import abstractmethod
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, TypeVar, Union

from docarray import Document, DocumentArray
from docarray.math import ndarray
from docarray.score import NamedScore
from qdrant_client.http import models as rest
from qdrant_client.http.models.models import Distance

if TYPE_CHECKING:  # pragma: no cover
    import numpy as np
    import tensorflow
    import torch
    from qdrant_client import QdrantClient

    QdrantArrayType = TypeVar(
        'QdrantArrayType',
        np.ndarray,
        tensorflow.Tensor,
        torch.Tensor,
        Sequence[float],
    )


class FindMixin:
    @property
    @abstractmethod
    def client(self) -> 'QdrantClient':
        raise NotImplementedError()

    @property
    @abstractmethod
    def collection_name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def serialize_config(self) -> dict:
        raise NotImplementedError()

    @property
    @abstractmethod
    def distance(self) -> 'Distance':
        raise NotImplementedError()

    def _find_similar_vectors(
        self,
        q: 'QdrantArrayType',
        limit: int = 10,
        filter: Optional[Dict] = None,
        search_params: Optional[Dict] = None,
        **kwargs,
    ):
        query_vector = self._map_embedding(q)

        search_result = self.client.search(
            self.collection_name,
            query_vector=query_vector,
            query_filter=filter,
            search_params=None
            if not search_params
            else rest.SearchParams(**search_params),
            limit=limit,
            append_payload=['_serialized'],
        )

        docs = []

        for hit in search_result:
            doc = Document.from_base64(
                hit.payload['_serialized'], **self.serialize_config
            )
            doc.scores[f'{self.distance.lower()}_similarity'] = NamedScore(
                value=hit.score
            )
            docs.append(doc)

        return DocumentArray(docs)

    def _find(
        self,
        query: 'QdrantArrayType',
        limit: int = 10,
        filter: Optional[Dict] = None,
        search_params: Optional[Dict] = None,
        **kwargs,
    ) -> List['DocumentArray']:
        """Returns approximate nearest neighbors given a batch of input queries.
        :param query: input supported to be used in Qdrant.
        :param limit: number of retrieved items
        :param filter: filter query used for pre-filtering
        :param search_params: additional parameters of the search


        :return: a list of DocumentArrays containing the closest Document objects for each of the queries in `query`.
        """

        num_rows, _ = ndarray.get_array_rows(query)

        if num_rows == 1:
            return [
                self._find_similar_vectors(
                    query, limit=limit, filter=filter, search_params=search_params
                )
            ]
        else:
            closest_docs = []
            for q in query:
                da = self._find_similar_vectors(
                    q, limit=limit, filter=filter, search_params=search_params
                )
                closest_docs.append(da)
            return closest_docs

    def _find_with_filter(
        self, filter: Optional[Dict], limit: Optional[Union[int, float]] = 10
    ):
        list_of_points, _offset = self.client.scroll(
            collection_name=self.collection_name,
            scroll_filter=rest.Filter(**filter),
            with_payload=True,
            limit=limit,
        )
        da = DocumentArray()
        for result in list_of_points[:limit]:
            doc = Document.from_base64(
                result.payload['_serialized'], **self.serialize_config
            )
            da.append(doc)
        return da

    def _filter(
        self, filter: Optional[Dict], limit: Optional[Union[int, float]] = 10
    ) -> 'DocumentArray':
        """Returns a subset of documents by filtering by the given filter (`Qdrant` filter)..
        :param limit: number of retrieved items
        :param filter: filter query used for filtering.
        For more information: https://docarray.jina.ai/advanced/document-store/qdrant/#qdrant
        :return: a `DocumentArray` containing the `Document` objects that verify the filter.
        """

        return self._find_with_filter(filter, limit=limit)
